import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { Inject } from '@nestjs/common';
import {
  AggregateID,
  ConflictException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import { CreateAuthenticationCommand } from './create-authentication.command';
import { AUTHENTICATION_REPOSITORY } from '@modules/authentication/authentication.di-tokens';
import { AuthenticationRepositoryPort } from '../../ports/authentication.repository.port';
import { AuthenticationEntity } from '@modules/authentication/core/domain/authentication.entity';
import {
  AuthenticationAlreadyExistsException,
  UsernameAlreadyExistsException,
} from '@modules/authentication/core/domain/authentication.errors';

@CommandHandler(CreateAuthenticationCommand)
export class CreateAuthenticationService implements ICommandHandler {
  constructor(
    @Inject(AUTHENTICATION_REPOSITORY)
    private readonly authenticationRepository: AuthenticationRepositoryPort,
  ) {}

  async execute(command: CreateAuthenticationCommand): Promise<AggregateID> {
    const authentication: AuthenticationEntity =
      await AuthenticationEntity.create({
        userId: command.userId,
        password: command.password,
        usernames: command.usernames,
      });
    try {
      await this.authenticationRepository.insert(authentication);
      return authentication.getProps().userId;
    } catch (error: any) {
      if (error instanceof ConflictException) {
        throw new AuthenticationAlreadyExistsException(error);
      }
      if (
        error instanceof UniqueConstraintException &&
        error.message.includes('uuid')
      ) {
        throw new AuthenticationAlreadyExistsException(error);
      }
      if (
        error instanceof UniqueConstraintException &&
        error.message.includes('username')
      ) {
        throw new UsernameAlreadyExistsException(error);
      }
      throw error;
    }
  }
}
