import {
  AggregateID,
  ConflictException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import { USERNAME_REPOSITORY } from '@modules/authentication/authentication.di-tokens';
import { UpdateUsernameCommand } from '@modules/authentication/core/application/commands/update-username/update-username.command';
import { UpdateUsernameService } from '@modules/authentication/core/application/commands/update-username/update-username.service';
import { UsernameAlreadyExistsException } from '@modules/authentication/core/domain/authentication.errors';
import { Type } from '@modules/authentication/core/domain/username.types';
import { UpdateUsernameRequestDto } from '@modules/authentication/interface/grpc-controllers/dtos/update-username.request.dto';
import { Test, TestingModule } from '@nestjs/testing';

const updateUsernameRequest: UpdateUsernameRequestDto = {
  userId: '165192d4-398a-4469-a16b-98c02cc6f531',
  name: 'new-john.doe@email.com',
  type: Type.EMAIL,
};

const mockUsernameRepository = {
  findByType: jest.fn().mockImplementation(() => ({
    id: 'john.doe@email.com',
    update: jest.fn(),
    getProps: jest.fn().mockImplementation(() => ({
      userId: '165192d4-398a-4469-a16b-98c02cc6f531',
      name: 'new-john.doe@email.com',
      type: Type.EMAIL,
    })),
  })),
  updateUsername: jest
    .fn()
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => {
      throw new ConflictException('already exists');
    })
    .mockImplementationOnce(() => {
      throw new UniqueConstraintException('username');
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Update Username Service', () => {
  let updateUsernameService: UpdateUsernameService;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: USERNAME_REPOSITORY,
          useValue: mockUsernameRepository,
        },
        UpdateUsernameService,
      ],
    }).compile();

    updateUsernameService = module.get<UpdateUsernameService>(
      UpdateUsernameService,
    );
  });

  it('should be defined', () => {
    expect(updateUsernameService).toBeDefined();
  });

  describe('execution', () => {
    const updateUsernameCommand = new UpdateUsernameCommand({
      userId: updateUsernameRequest.userId,
      username: {
        name: updateUsernameRequest.name,
        type: updateUsernameRequest.type,
      },
    });
    it('should update a username', async () => {
      const result: AggregateID = await updateUsernameService.execute(
        updateUsernameCommand,
      );
      expect(result).toBe('new-john.doe@email.com');
    });
    it('should throw a dedicated exception if username already exists for this type', async () => {
      await expect(
        updateUsernameService.execute(updateUsernameCommand),
      ).rejects.toBeInstanceOf(UsernameAlreadyExistsException);
    });
    it('should throw a dedicated exception if username already exists for this name', async () => {
      await expect(
        updateUsernameService.execute(updateUsernameCommand),
      ).rejects.toBeInstanceOf(UsernameAlreadyExistsException);
    });
    it('should throw an error if something bad happens', async () => {
      await expect(
        updateUsernameService.execute(updateUsernameCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });
});
